package ru.langrafhomyak.db.jdbc_resources_manager;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import ru.landrafhomyak.utility.reference_counter.CloseableReferenceCounter;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.net.URL;
import java.sql.Array;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.Date;
import java.sql.NClob;
import java.sql.ParameterMetaData;
import java.sql.PreparedStatement;
import java.sql.Ref;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.RowId;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.SQLXML;
import java.sql.Statement;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.Map;
import java.util.Objects;

/* package */ abstract class _PreparedStatementWrapper implements PreparedStatement {
	@NotNull
	protected final PreparedStatement _orig;
	@NotNull
	private final CloseableReferenceCounter _refcnt;
	@NotNull
	private final Object _sync;
	private ResultSet _currentQuery = null;
	private boolean _closeOnCompletion = false;

	_PreparedStatementWrapper(@NotNull PreparedStatement orig) {
		this._orig = orig;
		this._refcnt = new CloseableReferenceCounter("This prepared statement was returned to pool to be used in future");
		this._sync = new Object();
	}

	protected abstract void _onClose() throws SQLException;

	@Override
	public void close() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.close("Can't close prepared statement while it's in use");
			this._onClose();
		}
	}

	@Override
	public boolean isClosed() {
		return this._refcnt.isClosed();
	}

	@Override
	public void closeOnCompletion() {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._closeOnCompletion = true;
		}
	}

	@Override
	public boolean isCloseOnCompletion() {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._closeOnCompletion;
		}
	}

	private class _BoundResultSetWrapper extends ResultSetWrapper {

		public _BoundResultSetWrapper(@NotNull ResultSet orig) {
			super(orig);
		}

		@Override
		public void close() throws SQLException {
			if (this._orig.isClosed())
				return;
			this._orig.close();
			_PreparedStatementWrapper.this._currentQuery = null;
			if (_PreparedStatementWrapper.this._closeOnCompletion)
				_PreparedStatementWrapper.this.close();
		}

		@Override
		public Statement getStatement() {
			return _PreparedStatementWrapper.this;
		}

	}

	@NotNull
	private ResultSet _prepareRS(@Nullable ResultSet orig) {
		Objects.requireNonNull(orig, "Underlying PreparedStatement didn't returned ResultSet");
		ResultSet wrapped = this.new _BoundResultSetWrapper(orig);
		this._currentQuery = wrapped;
		return wrapped;
	}

	@Override
	public ResultSet executeQuery() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._prepareRS(this._orig.executeQuery());
		}
	}

	@Override
	public ResultSet getResultSet() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._prepareRS(this._orig.getResultSet());
		}
	}

	@Override
	public ResultSet getGeneratedKeys() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._prepareRS(this._orig.getGeneratedKeys());
		}
	}


	@Override
	public <T> T unwrap(java.lang.Class<T> iface) throws SQLException {
		return iface.isInstance(this) ? iface.cast(this) : this._orig.unwrap(iface);
	}

	@Override
	public boolean isWrapperFor(java.lang.Class<?> iface) throws SQLException {
		return iface.isInstance(this) || this._orig.isWrapperFor(iface);
	}


	@Override
	public void setPoolable(boolean value) throws SQLException {
		throw new SQLException("Disabling pooling not allowed on prepared statements driven by PreparedStatementsCompilationCache");

	}

	@Override
	public boolean isPoolable() {
		return true;
	}


	@Override
	public void cancel() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.cancel();
		}
	}

	@Override
	public int executeUpdate() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.executeUpdate();
		}
	}

	@Override
	public void setNull(int parameterIndex, int sqlType) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setNull(parameterIndex, sqlType);
		}
	}

	@Override
	public void setBoolean(int parameterIndex, boolean x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setBoolean(parameterIndex, x);
		}
	}

	@Override
	public void setByte(int parameterIndex, byte x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setByte(parameterIndex, x);
		}
	}

	@Override
	public void setShort(int parameterIndex, short x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setShort(parameterIndex, x);
		}
	}

	@Override
	public void setInt(int parameterIndex, int x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setInt(parameterIndex, x);
		}
	}

	@Override
	public void setLong(int parameterIndex, long x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setLong(parameterIndex, x);
		}
	}

	@Override
	public void setFloat(int parameterIndex, float x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setFloat(parameterIndex, x);
		}
	}

	@Override
	public void setDouble(int parameterIndex, double x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setDouble(parameterIndex, x);
		}
	}

	@Override
	public void setBigDecimal(int parameterIndex, BigDecimal x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setBigDecimal(parameterIndex, x);
		}
	}

	@Override
	public void setString(int parameterIndex, String x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setString(parameterIndex, x);
		}
	}

	@Override
	public void setBytes(int parameterIndex, byte[] x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setBytes(parameterIndex, x);
		}
	}

	@Override
	public void setDate(int parameterIndex, Date x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setDate(parameterIndex, x);
		}
	}

	@Override
	public void setTime(int parameterIndex, Time x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setTime(parameterIndex, x);
		}
	}

	@Override
	public void setTimestamp(int parameterIndex, Timestamp x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setTimestamp(parameterIndex, x);
		}
	}

	@Override
	public void setAsciiStream(int parameterIndex, InputStream x, int length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setAsciiStream(parameterIndex, x, length);
		}
	}

	@Override
	@SuppressWarnings({"deprecation"})
	public void setUnicodeStream(int parameterIndex, InputStream x, int length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setUnicodeStream(parameterIndex, x, length);
		}
	}

	@Override
	public void setBinaryStream(int parameterIndex, InputStream x, int length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setBinaryStream(parameterIndex, x);
		}
	}

	@Override
	public void clearParameters() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.clearParameters();
		}
	}

	@Override
	public void setObject(int parameterIndex, Object x, int targetSqlType) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setObject(parameterIndex, x, targetSqlType);
		}
	}

	@Override
	public void setObject(int parameterIndex, Object x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setObject(parameterIndex, x);
		}
	}

	@Override
	public boolean execute() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.execute();
		}
	}

	@Override
	public void addBatch() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.addBatch();
		}
	}

	@Override
	public void setCharacterStream(int parameterIndex, Reader reader, int length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setCharacterStream(parameterIndex, reader, length);
		}
	}

	@Override
	public void setRef(int parameterIndex, Ref x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setRef(parameterIndex, x);
		}
	}

	@Override
	public void setBlob(int parameterIndex, Blob x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setBlob(parameterIndex, x);
		}
	}

	@Override
	public void setClob(int parameterIndex, Clob x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setClob(parameterIndex, x);
		}
	}

	@Override
	public void setArray(int parameterIndex, Array x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setArray(parameterIndex, x);
		}
	}

	@Override
	public ResultSetMetaData getMetaData() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getMetaData();
		}
	}

	@Override
	public void setDate(int parameterIndex, Date x, Calendar cal) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setDate(parameterIndex, x);
		}
	}

	@Override
	public void setTime(int parameterIndex, Time x, Calendar cal) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setTime(parameterIndex, x);
		}
	}

	@Override
	public void setTimestamp(int parameterIndex, Timestamp x, Calendar cal) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setTimestamp(parameterIndex, x);
		}
	}

	@Override
	public void setNull(int parameterIndex, int sqlType, String typeName) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setNull(parameterIndex, sqlType, typeName);
		}
	}

	@Override
	public void setURL(int parameterIndex, URL x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setURL(parameterIndex, x);
		}
	}

	@Override
	public ParameterMetaData getParameterMetaData() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getParameterMetaData();
		}
	}

	@Override
	public void setRowId(int parameterIndex, RowId x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setRowId(parameterIndex, x);
		}
	}

	@Override
	public void setNString(int parameterIndex, String value) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setNString(parameterIndex, value);
		}
	}

	@Override
	public void setNCharacterStream(int parameterIndex, Reader value, long length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setNCharacterStream(parameterIndex, value, length);
		}
	}

	@Override
	public void setNClob(int parameterIndex, NClob value) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setNClob(parameterIndex, value);
		}
	}

	@Override
	public void setClob(int parameterIndex, Reader reader, long length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setClob(parameterIndex, reader, length);
		}
	}

	@Override
	public void setBlob(int parameterIndex, InputStream inputStream, long length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setBlob(parameterIndex, inputStream, length);
		}
	}

	@Override
	public void setNClob(int parameterIndex, Reader reader, long length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setNClob(parameterIndex, reader, length);
		}
	}

	@Override
	public void setSQLXML(int parameterIndex, SQLXML xmlObject) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setSQLXML(parameterIndex, xmlObject);
		}
	}

	@Override
	public void setObject(int parameterIndex, Object x, int targetSqlType, int scaleOrLength) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setObject(parameterIndex, x, targetSqlType, scaleOrLength);
		}
	}

	@Override
	public void setAsciiStream(int parameterIndex, InputStream x, long length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setAsciiStream(parameterIndex, x, length);
		}
	}

	@Override
	public void setBinaryStream(int parameterIndex, InputStream x, long length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setBinaryStream(parameterIndex, x, length);
		}
	}

	@Override
	public void setCharacterStream(int parameterIndex, Reader reader, long length) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setCharacterStream(parameterIndex, reader, length);
		}
	}

	@Override
	public void setAsciiStream(int parameterIndex, InputStream x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setAsciiStream(parameterIndex, x);
		}
	}

	@Override
	public void setBinaryStream(int parameterIndex, InputStream x) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setBinaryStream(parameterIndex, x);
		}
	}

	@Override
	public void setCharacterStream(int parameterIndex, Reader reader) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setCharacterStream(parameterIndex, reader);
		}
	}

	@Override
	public void setNCharacterStream(int parameterIndex, Reader value) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setNCharacterStream(parameterIndex, value);
		}
	}

	@Override
	public void setClob(int parameterIndex, Reader reader) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setClob(parameterIndex, reader);
		}
	}

	@Override
	public void setBlob(int parameterIndex, InputStream inputStream) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setBlob(parameterIndex, inputStream);
		}
	}

	@Override
	public void setNClob(int parameterIndex, Reader reader) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setNClob(parameterIndex, reader);
		}
	}

	@Override
	public ResultSet executeQuery(String sql) throws SQLException {
		throw new SQLException("executeQuery(String) not allowed on PreparedStatement");
	}

	@Override
	public int executeUpdate(String sql) throws SQLException {
		throw new SQLException("executeUpdate(String) not allowed on PreparedStatement");
	}

	@Override
	public int getMaxFieldSize() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getMaxFieldSize();
		}
	}

	@Override
	public void setMaxFieldSize(int max) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setMaxFieldSize(max);
		}
	}

	@Override
	public int getMaxRows() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getMaxRows();
		}
	}

	@Override
	public void setMaxRows(int max) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setMaxRows(max);
		}
	}

	@Override
	public void setEscapeProcessing(boolean enable) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setEscapeProcessing(enable);
		}
	}

	@Override
	public int getQueryTimeout() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getQueryTimeout();
		}
	}

	@Override
	public void setQueryTimeout(int seconds) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setQueryTimeout(seconds);
		}
	}

	@Override
	public SQLWarning getWarnings() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getWarnings();
		}
	}

	@Override
	public void clearWarnings() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.clearWarnings();
		}
	}

	@Override
	public void setCursorName(String name) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setCursorName(name);
		}
	}


	@Override
	public boolean execute(String sql) throws SQLException {
		throw new SQLException("execute(String) not allowed on PreparedStatement");
	}


	@Override
	public int getUpdateCount() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getUpdateCount();
		}
	}

	@Override
	public boolean getMoreResults() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getMoreResults();
		}
	}

	@Override
	public void setFetchDirection(int direction) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setFetchDirection(direction);
		}
	}

	@Override
	public int getFetchDirection() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getFetchDirection();
		}
	}

	@Override
	public void setFetchSize(int rows) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.setFetchSize(rows);
		}
	}

	@Override
	public int getFetchSize() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getFetchSize();
		}
	}

	@Override
	public int getResultSetConcurrency() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getResultSetConcurrency();
		}
	}

	@Override
	public int getResultSetType() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getResultSetType();
		}
	}

	@Override
	public void addBatch(String sql) throws SQLException {
		throw new SQLException("addBatch(String) not allowed on PreparedStatement");
	}

	@Override
	public void clearBatch() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			this._orig.clearBatch();
		}
	}

	@Override
	public int[] executeBatch() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.executeBatch();
		}
	}

	@Override
	public Connection getConnection() throws SQLException {
		return this._orig.getConnection();
	}

	@Override
	public boolean getMoreResults(int current) throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getMoreResults(current);
		}
	}

	@Override
	public int executeUpdate(String sql, int autoGeneratedKeys) throws SQLException {
		throw new SQLException("executeUpdate(String, int) not allowed on PreparedStatement");
	}

	@Override
	public int executeUpdate(String sql, int[] columnIndexes) throws SQLException {
		throw new SQLException("executeUpdate(String, int[]) not allowed on PreparedStatement");
	}

	@Override
	public int executeUpdate(String sql, String[] columnIndexes) throws SQLException {
		throw new SQLException("executeUpdate(String, String[]) not allowed on PreparedStatement");
	}

	@Override
	public boolean execute(String sql, int autoGeneratedKeys) throws SQLException {
		throw new SQLException("executeUpdate(String, int) not allowed on PreparedStatement");
	}

	@Override
	public boolean execute(String sql, int[] columnIndexes) throws SQLException {
		throw new SQLException("executeUpdate(String, int[]) not allowed on PreparedStatement");
	}

	@Override
	public boolean execute(String sql, String[] columnIndexes) throws SQLException {
		throw new SQLException("executeUpdate(String, String[]) not allowed on PreparedStatement");
	}

	@Override
	public int getResultSetHoldability() throws SQLException {
		synchronized (this._sync) {
			this._refcnt.assertNotClosed();
			return this._orig.getResultSetHoldability();
		}
	}
}
